###
### TCGUI
### Run time module

if {[info procs cWidget] != ""} {
	return
}

############## Variables ###################

proc set_tcgui {ix lst} {
	global tcgui
	foreach {name val} $lst {
		set tcgui($ix,$name) $val
	}
}

set_tcgui cmds { DISPLAY 1 WITHDRAW 2 SET-ATTR 3 INSERT-VALUE 4 
	SELECT-VALUE 5 WAIT 6}
set_tcgui events { ERROR 1 CLOSE 2 BUTTON 3 }
set_tcgui tkattb { 1 foreground 2 background 3 font 4 relief 5 border }

# Widget classes: Label Entry Group Image ComboBox 
# Listbox Button Menubutton Checkbutton Radiobutton Text

set tcgui(tcl_attribs) {foreground background font relief border underline}
set_tcgui attrib { 
	Label { label foreground background font relief border}
	Entry { cobolvar cobsize foreground background font 
		relief border }
	Group { label cobolvar cobsize foreground background 
		font relief border }
	Image { cobolvar cobsize background relief border}
	ComboBox { cobolvar cobsize foreground background font 
		relief border }
	Listbox { cobolvar cobsize foreground background font 
		relief border }
	Button { label foreground background font relief border }
	Menuitem { widgetname label underline eventname cascade }
	Menubutton { label underline foreground background font relief border }
	Checkbutton { label cobolvar cobsize  
		onvalue offvalue foreground background font relief border}
	Radiobutton { label cobolvar cobsize selvalue foreground 
		background font relief border }
	Text { cobolvar cobsize foreground background font 
		relief border }
}

############## TinyCobol <--> tcl/tk communications #################

proc get_cobvars {} {
	global tc tcgui ready
	set appl [string tolower $tc(gui_name)]
	set wappl .$appl
	set tc(data_ptr,$appl) $tc(data_ptr)
	switch $tc(cmd) {
	1 {
		wm deiconify $wappl
		foreach {ix name} [array get tcgui cobolvar,$wappl.*] {
			if {$name != ""} {
				set w [lindex [split $ix ,] 1]
				set size $tcgui(cobsize,$w)
				getcobvar val $tcgui(coboffset,$w) $size
				switch [winfo class $w] {
				Entry {
					$w delete 0 end	
					$w insert end [string trimright $val]
				}
				Text {
					$w config -state normal
					$w delete 0.0 end	
					$w insert end [string trimright $val] center
					$w config -state disabled 
				}
				Radiobutton {
					set varname $tcgui(cobolvar,$w)
					#set tc($appl,$varname) $tcgui(value,$w,selval)
					getcobvar tc($appl,$varname) $tcgui(coboffset,$w) 1
				}
				Checkbutton {
					set varname $tcgui(cobolvar,$w)
					getcobvar tc(check,$appl,$varname) $tcgui(coboffset,$w) 1
				}
				Image {
					if {[info exists tc(image,$w)]} {
						$tc(image,$app,$w) destroy
					}
					set tc(image,$appl,$w) [image create photo \
						-file [string trimright $val]]
					$w.i config -image $tc(image,$appl,$w)
				}
				}
			}
		}
		next_taborder $wappl
	}
	2 {
		wm withdraw $wappl
	}
	3 {
		set wname [format %03d $tc(widget_sel)]
		set w $tc($wappl,$wname)
		set_attribs $w.w $tcgui(tkattb,$tc(attr_type)) $tc(attr_value) 
	}
	4 {
		set wname [format %03d $tc(widget_sel)]
		foreach {ix name} [array get tcgui cobolvar,$wappl.*$wname*] {
			if {$name != ""} {
				set w [lindex [split $ix ,] 1]
				set size $tcgui(cobsize,$w)
				getcobvar val $tcgui(coboffset,$w) $size
				if {[winfo class $w] == "ComboBox"} {
					$w.btm.lbox insert end [string trimright $val]
				}
				if {[winfo class $w] == "Listbox"} {
					$w insert end $val
				}
			}
		}
		set tc(event) 0
	}
	5 {
		set wname [format %03d $tc(widget_sel)]
		foreach {ix name} [array get tcgui cobolvar,$wappl.*$wname*] {
			if {$name != ""} {
				set w [lindex [split $ix ,] 1]
				set size $tcgui(cobsize,$w)
				getcobvar val $tcgui(coboffset,$w) $size
				if {[winfo class $w] == "ComboBox"} {
					$w.e delete 0 end
					$w.e insert end [string trimright $val]
				}
			}
		}
		set tc(event) 0
	}
	6 {
		wm deiconify $wappl
		tkwait variable ready
	}
	}
}

proc send_results {appl} {
	global tc tcgui
	set appl [string tolower $appl]
	foreach {ix name} [array get tcgui cobolvar,.$appl.*] {
		if {$name != ""} {
			set w [lindex [split $ix ,] 1]
			set size $tcgui(cobsize,$w)
			set uval ""
			switch [winfo class $w] {
			Entry { set uval [$w get] }
			Text { set uval [$w get 0.0 end] }
		    Listbox { 
					set ix [$w curselection]
					if {$ix != ""} {
						set uval [$w get $ix]
					}
				}
			ComboBox { set uval [$w.e get] }
			}
			if {$uval != ""} {
				set val [format "%-$size.${size}s" $uval]
				setcobvar $val $tcgui(coboffset,$w) $size $tc(data_ptr)
			}
		}
	}
	# set variables modified by radiobuttons
	foreach {ix val} [array get tc $appl,*] {
		set varname [lindex [split $ix ,] 1]
		set val [string index $tc($ix) 0]
		setcobvar $val $tc(offset,$ix) 1 $tc(data_ptr)
	}
	# set variables modified by checkbuttons
	foreach {ix val} [array get tc check,$appl,*] {
		set varname [lindex [split $ix ,] 2]
		set ix1 [join [lrange [split $ix ,] 1 end] ,]
		setcobvar $tc($ix) $tc(offset,$ix1) 1 $tc(data_ptr)
	}
}

proc butevent {ev appl} {
	global tc ready
	set appl [string tolower $appl]
	set tc(gui_name) $appl
	set tc(data_ptr) $tc(data_ptr,$appl)
	send_results $appl
	set tc(event) $ev
	set ready 1
}

proc set_mainwin {appl width height} {
	global tcgui mainwin
	set w .[string tolower $appl]
	set mainwin [toplevel $w -width $width -height $height]
	wm withdraw $w 
	set tcgui(usr_ev) 10
	set tcgui(appl) $appl
}

wm withdraw .
bind all <Control-c> exit
bind all <Control-d> {after 100 set ready 1 ; break}
bind edit <Tab> {next_taborder %W ; break}
bind edit <Alt-Tab> {next_toplevel %W ; break}
bind edit <Shift-Tab> {prev_taborder %W ; break}

proc set_taborder {args} {
	global tcgui tc
	set appl [string tolower $tcgui(appl)]
	set new {}
	foreach n $args {
		lappend new [lindex $tc(taborder,$appl) \
			[lsearch -glob $tc(taborder,$appl) *$n*]]
	}
	set tc(taborder,$appl) $new
}

proc prev_taborder {w} {
	global tc tcgui
	set appl [lindex [split $w .] 1]
	set tb $tc(taborder,$appl)
	set n [lsearch $tb $w]
	if {$n > 0} {
		incr n -1
	} else {
		set n [expr [llength $tb]-1]
	}
	if {$n < 0} {
		focus [lindex $tb end]
	} else {
		focus [lindex $tb $n]
	}
}

proc next_toplevel {w} {
	global tc tcgui
	set appl [lindex [split $w .] 1]
	set tops [winfo children .]
	for {set i 0} {$i < [llength $tops]} {incr i} {
		set t [lindex $tops $i]
		if {![winfo ismapped $t]} {
			set tops [lreplace $tops $i $i]
		}	
	}
	set i [lsearch $tops .$appl]
	if {$i < 0} {
		set i 0
	} else {
		incr i
		if {$i >= [llength $tops]} {
			set i 0
		}
	}
	set appl [string range [lindex $tops $i] 1 end]
	set tb $tc(taborder,$appl)
	focus [lindex $tb 0]
}

proc next_taborder {w} {
	global tc tcgui
	set appl [lindex [split $w .] 1]
	set tb $tc(taborder,$appl)
	set n [lsearch $tb $w]
	if {$n >= 0} {
		incr n
	} else {
		set n 0
	}
	if {$n < [llength $tb]} {
		focus [lindex $tb $n]
	} else {
		focus [lindex $tb 0]
	}
}

proc cWidget { type name parent x y wd ht var size offset args } {
	global tcgui mainwin tc 
	if {$parent == "."} {
		set parent $mainwin 
	} else {
		set parent $mainwin$parent
	}
	set appl [string tolower $tcgui(appl)]
	if {![info exists tcgui(taborder,$appl)]} {
		set tcgui(taborder,$appl) {}
	}
	if {$type == "cMenuitem"} {
		set w $parent.m
	} else {
		set w [frame [join [list $parent $name] .]]
		if {$wd} {
			$w config -width $wd
		}
		if {$ht} {
			$w config -height $ht
		}
	}
	switch $type {
	cLabel { 
			set ww [label $w.w ]
		}
	cEntry {
			set ww [entry $w.w -width $size -textvariable tc($mainwin,$var)]
			lappend tc(taborder,$appl) $w.w
		}
	cList {
			scrollbar $w.sv -orient vertical -width 8 \
				-border 1 -command [list $w.w yview] -takefocus 0
			scrollbar $w.sh -orient horizontal -width 8 \
				-border 1 -command [list $w.w xview] -takefocus 0
			set ww [listbox $w.w -font {courier 12} -width $size \
				-yscrollcommand [list $w.sv set] \
				-xscrollcommand [list $w.sh set]]
			pack $w.sh -side bottom -fill x
			pack $w.sv -side right -fill y
			lappend tc(taborder,$appl) $w.w
		}
	cButton {
			set ww [button $w.w \
				-command [list butevent [incr tcgui(usr_ev)] $tcgui(appl)]]
			lappend tc(taborder,$appl) $w.w
		}
	cMenubutton {
			set ww [menubutton $w.w -text Menu -underline 0 \
				-menu $w.w.m -relief flat]
			menu $w.w.m -tearoff 0
		}
	cMenuitem {
			set ww {}
			if {[winfo class $parent] == "Menu"} {
				set m $parent
			} else {
				set m $parent.m
			}
			set mi [$m index end]
			if {$mi == "none"} {
				set mi 0
			} else {
				incr mi
			}
			array set wargs $args
			if {[info exists wargs(cascade)] &&
				$wargs(cascade) == 1} {
					$m add cascade -menu $m.$mi
					menu $m.$mi -tearoff 0
			} elseif {[lsearch -exact $args "separator"] >= 0} {
				$m add separator
			} else {
				$m add command -command \
					[list butevent [incr tcgui(usr_ev)] $tcgui(appl)]
			}
		}
	cCheck {
			set ww [checkbutton $w.w]
			lappend tc(taborder,$appl) $w.w
		}
	cRadio {
			set ww [radiobutton $w.w]
			lappend tc(taborder,$appl) $w.w
		}
	cGroup {
			set ww [frame $w.w -class Group]
			pack [frame $w.w.fr -bd 2 -relief groove -width 100 -height 100] \
				-padx 10 -pady 10 -expand 1 -fill both
			place [label $w.w.lab] -in $w.w.fr -x 7 -y 0 -anchor w
			bindtags $w.w.fr [concat edit [bindtags $w.w.fr]]
		}
	cText { 
			set ww [text $w.w -bg white \
				-wrap word -state disabled]
			$w.w tag config center -justify center
		}
	cCombo { 
			set ww [combobox $w.w ]
			bindtags $w.w.e [concat edit [bindtags $w.w.e]]
			bindtags $w.w.b [concat edit [bindtags $w.w.b]]
			lappend tc(taborder,$appl) $w.w.e
		}
	cImage { 
			set ww [frame $w.w -class Image]
			pack [label $w.w.i] -expand 1 -fill both
		}
	}
	if {$ww == {}} {
		# menu items are not packed
		eval set_menu_attribs $m $mi $args
		return
	}
	pack $ww -expand 1 -fill both
	place $w -x $x -y $y
	update
	set wid [winfo width $w]
	set hei [winfo height $w]
	pack propagate $w 0
	$w config -height $hei -width $wid
	bindtags $w.w [concat edit [bindtags $w.w]]
	if {$wd} {
		$w config -width $wd
	}
	if {$ht} {
		$w config -height $ht
	}
	set tcgui(cobolvar,$w.w) $var
	set tcgui(cobsize,$w.w) $size
	set tcgui(coboffset,$w.w) $offset
	set tc($mainwin,$name) $w
	eval set_attribs $w.w $args
	return $w	
}

proc cLabel {args} { 	eval cWidget cLabel $args }
proc cEntry {args} { 	eval cWidget cEntry $args }
proc cList {args} { 	eval cWidget cList $args }
proc cButton {args} { 	eval cWidget cButton $args }
proc cMenubutton {args} { 	eval cWidget cMenubutton $args }
proc cMenuitem {args} { 	eval cWidget cMenuitem $args }
proc cCheck {args} { 	eval cWidget cCheck $args }
proc cRadio {args} { 	eval cWidget cRadio $args }
proc cGroup {args} { 	eval cWidget cGroup $args }
proc cText {args} { 	eval cWidget cText $args }
proc cCombo {args} { 	eval cWidget cCombo $args }
proc cImage {args} { 	eval cWidget cImage $args }

################ WIDGET PROPERTIES DIALOGS ################
proc apply_attrib {w {mitem {}}} {
	global tcgui tc
	set class [winfo class $w]
	if {$class == "Group"} {
		foreach at $tcgui(attrib,$class) {
			if {[info exists tcgui(value,$w,$at)] &&
				$tcgui(value,$w,$at) != "" &&
				[lsearch $tcgui(tcl_attribs) $at]>=0} {
					$w.fr config -$at $tcgui(value,$w,$at)
					if {[lsearch {border relief} $at]<0} {
						$w.lab config -$at $tcgui(value,$w,$at)
					}
			} elseif {$at == "label" && 
				[info exists tcgui(value,$w,label)]} {
			# special label attribute for buttons, groups and labels
					$w.lab config -text $tcgui(value,$w,$at)
			}
		}	
	} elseif {$class == "ComboBox"} {
		foreach at $tcgui(attrib,$class) {
			# standard tcl attributes
			if {[info exists tcgui(value,$w,$at)] &&
				$tcgui(value,$w,$at) != "" &&
				[lsearch $tcgui(tcl_attribs) $at]>=0} {
					$w.e config -$at $tcgui(value,$w,$at)
			} elseif {$at == "label" &&
				[info exists tcgui(value,$w,label)]} {
			# special label attribute for buttons, groups and labels
				$w.e config -text $tcgui(value,$w,$at)
			} 
		}
	} elseif {$class == "Menu"} {
		foreach at $tcgui(attrib,Menuitem) {
#			if {$tcgui(value,$w,$mitem,cascade) == 1 &&
#				[$w type $mitem] != "cascade"} {
#				$w delete $mitem 
#				$w add cascade -menu $w.$mitem
#				destroy $w.$mitem
#				menu $w.$mitem -tearoff 0
#			}
			if {[info exists tcgui(value,$w,$mitem,$at)] &&
				$tcgui(value,$w,$mitem,$at) != "" &&
				[lsearch $tcgui(tcl_attribs) $at]>=0} {
					$w entryconfig $mitem -$at $tcgui(value,$w,$mitem,$at)
			} elseif {$at == "label" &&
				[info exists tcgui(value,$w,$mitem,label)]} {
				$w entryconfig $mitem -label $tcgui(value,$w,$mitem,$at)
			}
		}
	} elseif {$class == "Radiobutton"} {
		foreach at $tcgui(attrib,$class) {
			# standard tcl attributes
			if {[info exists tcgui(value,$w,$at)] &&
				$tcgui(value,$w,$at) != "" &&
				[lsearch $tcgui(tcl_attribs) $at]>=0} {
					$w config -$at $tcgui(value,$w,$at)
			} elseif {$at == "label" &&
				[info exists tcgui(value,$w,label)]} {
				$w config -text $tcgui(value,$w,$at)
			} elseif {$at == "cobolvar"} {
				set appl [string tolower $tc(gui_name)]
				set varname $tcgui(cobolvar,$w)
				$w config -variable tc($appl,$varname) \
					-value $tcgui(value,$w,selval)
				set tc(offset,$appl,$varname) $tcgui(coboffset,$w)
			}
		}
	} elseif {$class == "Checkbutton"} {
		foreach at $tcgui(attrib,$class) {
			# standard tcl attributes
			if {[info exists tcgui(value,$w,$at)] &&
				$tcgui(value,$w,$at) != "" &&
				[lsearch $tcgui(tcl_attribs) $at]>=0} {
					$w config -$at $tcgui(value,$w,$at)
			} elseif {$at == "label" &&
				[info exists tcgui(value,$w,label)]} {
				$w config -text $tcgui(value,$w,$at)
			} elseif {$at == "cobolvar"} {
				set appl [string tolower $tc(gui_name)]
				set varname $tcgui(cobolvar,$w)
				$w config -variable tc(check,$appl,$varname) 
				set tc(offset,$appl,$varname) $tcgui(coboffset,$w)
			}
		}
	} elseif {$class == "Image"} {
		foreach at $tcgui(attrib,$class) {
			# standard tcl attributes
			if {[info exists tcgui(value,$w,$at)] &&
				$tcgui(value,$w,$at) != "" &&
				[lsearch $tcgui(tcl_attribs) $at]>=0} {
					$w config -$at $tcgui(value,$w,$at)
			} 
		}
	} else {
		foreach at $tcgui(attrib,$class) {
			# standard tcl attributes
			if {[info exists tcgui(value,$w,$at)] &&
				$tcgui(value,$w,$at) != "" &&
				[lsearch $tcgui(tcl_attribs) $at]>=0} {
					$w config -$at $tcgui(value,$w,$at)
			} elseif {$at == "label" &&
				[info exists tcgui(value,$w,label)]} {
			# special label attribute for buttons, groups and labels
				$w config -text $tcgui(value,$w,$at)
			} 
		}
	}
}

proc set_menu_attribs {m mi args} {
	global tcgui
	foreach {at val} $args {
		set	tcgui(value,$m,$mi,$at) $val
	}
	apply_attrib $m $mi
}


proc set_attribs {w args} {
	global tcgui
	foreach {at val} $args {
		set	tcgui(value,$w,$at) $val
	}
	apply_attrib $w
}

###############  UTILITY ROUTINES ##########################
# base64.tcl --
#
# Encode/Decode base64 for a string
# Stephen Uhler / Brent Welch (c) 1997 Sun Microsystems
# The decoder was done for exmh by Chris Garrigues
#
# Copyright (c) 1998-2000 by Ajuba Solutions.
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# 
# RCS: @(#) $Id: base64.tcl,v 1.8 2000/10/11 20:04:11 welch Exp $

# Version 1.0 implemented Base64_Encode, Bae64_Decode
# Version 2.0 uses the base64 namespace
# Version 2.1 fixes various decode bugs and adds options to encode

namespace eval base64 {
    variable i 0
    variable char
    variable base64
    variable base64_en
    foreach char {A B C D E F G H I J K L M N O P Q R S T U V W X Y Z \
	      a b c d e f g h i j k l m n o p q r s t u v w x y z \
	      0 1 2 3 4 5 6 7 8 9 + /} {
	set base64($char) $i
	set base64_en($i) $char
	incr i
    }

    namespace export *
}

# base64::encode --
#
#	Base64 encode a given string.
#
# Arguments:
#	args	?-maxlen maxlen? ?-wrapchar wrapchar? string
#	
#		If maxlen is 0, the output is not wrapped.
#
# Results:
#	A Base64 encoded version of $string, wrapped at $maxlen characters
#	by $wrapchar.

proc base64::encode {args} {
    variable base64_en
    
    # Set the default wrapchar and maximum line length to match the output
    # of GNU uuencode 4.2.  Various RFC's allow for different wrapping 
    # characters and wraplengths, so these may be overridden by command line
    # options.
    set wrapchar "\n"
    set maxlen 60

    if { [llength $args] == 0 } {
	error "wrong # args: should be \"[lindex [info level 0] 0]\
		?-maxlen maxlen? ?-wrapchar wrapchar? string\""
    }

    set optionStrings [list "-maxlen" "-wrapchar"]
    for {set i 0} {$i < [llength $args] - 1} {incr i} {
	set arg [lindex $args $i]
	set index [lsearch -glob $optionStrings "${arg}*"]
	if { $index == -1 } {
	    error "unknown option \"$arg\": must be -maxlen or -wrapchar"
	}
	incr i
	if { $i >= [llength $args] - 1 } {
	    error "value for \"$arg\" missing"
	}
	set val [lindex $args $i]
	set [string range [lindex $optionStrings $index] 1 end] $val
    }
    
    if { ![string is integer -strict $maxlen] } {
	error "expected integer but got \"$maxlen\""
    }

    set string [lindex $args end]

    set result {}
    set state 0
    set length 0
    foreach {c} [split $string {}] {
	# Do the line length check before appending so that we don't get an
	# extra newline if the output is a multiple of $maxlen chars long.
	if {$maxlen && $length >= $maxlen} {
	    append result $wrapchar
	    set length 0
	}
	scan $c %c x
	switch [incr state] {
	    1 {	append result $base64_en([expr {($x >>2) & 0x3F}]) }
	    2 { append result \
		$base64_en([expr {(($old << 4) & 0x30) | (($x >> 4) & 0xF)}]) }
	    3 { append result \
		$base64_en([expr {(($old << 2) & 0x3C) | (($x >> 6) & 0x3)}])
		append result $base64_en([expr {($x & 0x3F)}])
		incr length
		set state 0}
	}
	set old $x
	incr length
    }
    set x 0
    switch $state {
	0 { # OK }
	1 { append result $base64_en([expr {(($old << 4) & 0x30)}])== }
	2 { append result $base64_en([expr {(($old << 2) & 0x3C)}])=  }
    }
    return $result
}

#
# base64::decode --
#
#	Base64 decode a given string.
#
# Arguments:
#	string	The string to decode.  Characters not in the base64
#		alphabet are ignored (e.g., newlines)
#
# Results:
#	The decoded value.

proc base64::decode {string} {
    variable base64

    set output {}
    set group 0
    set j 18
    foreach char [split $string {}] {
	if {[string compare $char "="]} {
	    # RFC 2045 says that line breaks and other characters not part
	    # of the Base64 alphabet must be ignored, and that the decoder
	    # can optionally emit a warning or reject the message.  We opt
	    # not to do so, but to just ignore the character.

	    if { ![info exists base64($char)] } {
		continue
	    }
	    set bits $base64($char)
	    set group [expr {$group | ($bits << $j)}]
	    if {[incr j -6] < 0} {
		scan [format %06x $group] %2x%2x%2x a b c
		append output [format %c%c%c $a $b $c]
		set group 0
		set j 18
	    }
	} else {
	    # = indicates end of data.  Output whatever chars are left.
	    # The encoding algorithm dictates that we can only have 1 or 2
	    # padding characters.  If j is 6, we have 12 bits of input 
	    # (enough for 1 8-bit output).  If j is 0, we have 18 bits of
	    # input (enough for 2 8-bit outputs).
	    # It is crucial to scan three hex digits even though we
	    # discard c - older code used %04x and scanned 2 hex digits
	    # but really ended up generating 5 or 6 (not 4!) digits and
	    # resulted in alignment errors.

	    scan [format %06x $group] %2x%2x%2x a b c
	    if {$j == 6} {
		append output [format %c $a]
	    } elseif {$j == 0} {
		append output [format %c%c $a $b]
	    }
	    break
	}
    }
    return $output
}

#package provide base64 2.1
namespace import base64::*

proc compare {s1 s2} {
    if {[string equal $s1 $s2]} {
	return equal
    }
    if {[string length $s1] != [string length $s2]} {
	return "s1 length [string length $s1] != s2 length [string length $s2]"
    }
    set i 0
    foreach c1 [split $s1 {}] c2 [split $s2 {}] {
	if {[string compare $c1 $c2] != 0} {
	    scan $c1 %c c1scanval
	    scan $c2 %c c2scanval
	    binary scan $c1 a c1binval
	    binary scan $c2 a c2binval
	    return "char($i) c1 $c1 $c1scanval $c1binval c2 $c2 $c2scanval $c2binval"
	}
	incr i
    }
}

############ EXECUTE ENCODED SOURCE ###############
proc exec_encoded {} {
	global encoded_source mainwin
	catch {file delete tmp.tcl.gz tmp.tcl}
	set f [open tmp.tcl.gz w]
	fconfigure $f -encoding binary -translation binary
	puts -nonewline $f [decode $encoded_source]
	close $f
	exec gunzip tmp.tcl.gz
	uplevel #0 source tmp.tcl
	catch {file delete tmp.tcl}
	flush stdout
}

# This is a (slightly) improved version of Jeff's vanilla Tk combobox
# (less problems with binding clashes, uses a bitmap on the button, a
# few other features...) but the real effort is his, so I:
#    a) include his original news posting (with as much in the way of
#       headers as my newsreader presents on verbose :^). This also
#       acts as documentation...
#    b) acknowledge that he is the prior author
#
# If you want to use this code commercially, you'll probably need to
# obtain permission from both of us (I imagine that I'd agree without
# any problems, but I can't speak for Jeffrey).
#
# Enjoy!
#
# Donal.  ( mailto:fellowsd@cs.man.ac.uk  -  http://r8h.cs.man.ac.uk:8000/ )

#comp.lang.tcl #32821 (0 + 394 more)
#Path: cs.man.ac.uk!yama.mcc.ac.uk!zippy.dct.ac.uk!btnet!tank.news.pipex.net!
#+     pipex!news.mathworks.com!newsfeed.internetmci.com!chi-news.cic.net!news.
#+     uoregon.edu!cs.uoregon.edu!cs.uoregon.edu!news
#Newsgroups: comp.lang.tcl
#Subject: CODE: Vanilla Tk4 ComboBox
#Message-ID: <46lr31$30t@psychotix.cs.uoregon.edu>
#From: jhobbs@cs.uoregon.edu (Jeffrey Hobbs)
#Date: 25 Oct 1995 10:10:57 -0700
#Organization: University of Oregon Computer and Information Sciences Dept.
#NNTP-Posting-Host: psychotix.cs.uoregon.edu
#Lines: 101
#
#Several people were asking about alternatives to the combobox if they
#didn't have Tix or something.  Since they are so useful, and relatively
#easy to make, I coded some *example* code for a vanilla Tk combobox.
#For the uninitiated, a combobox is just an entry box linked to a popup
#listbox that hold history and such.
#
#This is not a good replacement for Tix's tixComboBox or the incrTk2
#combobox, but is useful starting point for those extensionless users.
#It's mostly an example of an overrideredirected toplevel.
#
#Some notes: double-1 or <Escape> in the entry widget, or selecting the
#button will toggle the listbox portion.  <Escape> will close the listbox
#without a selection.  <Return> in the entry widget adds items (no
#duplicates) to the listbox.  <Tab> in the entry widget searches the listbox
#for a unique match.  <Double-1> in the listbox selects that item.  The
#listbox scrollbar appears when more than 5 items are in the listbox when it
#pops up.  Any args beyond the window name go directly to the entry widget.
#
#Example use:
#pack [combobox .combo]
#pack [combobox .combo -width 20 -textvariable myvar]
#
### BEGIN COMBOBOX CODE ----------8<----------8<-----------8<-----------

set bmdata "#define downbut_width 14\n#define downbut_height 14\n"
append bmdata {static char downbut_bits[] = {
   0x00, 0x00, 0xe0, 0x01, 0xe0, 0x01, 0xe0, 0x01, 0xe0, 0x01, 0xfc, 0x0f,
   0xf8, 0x07, 0xf0, 0x03, 0xe0, 0x01, 0xc0, 0x00, 0x00, 0x00, 0xfe, 0x1f,
   0xfe, 0x1f, 0x00, 0x00};
}
image create bitmap comboboxbuttonimage -data $bmdata 

bind ComboEntry <Double-1> {combobox_popup [winfo parent %W]}
bind ComboEntry <Down>   {combobox_popup [winfo parent %W]}
bind ComboEntry <Control-f>      {combobox_find  %W; break}
bind ComboEntry <Return>   {combobox_add   %W}

bind ComboList  <Escape>   {combobox_popup [winfo parent [winfo parent %W]]}
bind ComboList  <Double-1> {combobox_get   %W %y}
bind ComboList  <Return> {combobox_get   %W}

proc combobox {w args} {
    global comboboxbuttonimage
    if [winfo exists $w] { error "window $w already exists" }

    frame $w -class ComboBox

    ## Entry Widget, it gets all the args
    eval entry $w.e $args

    ## Button.  With the symbol font, it should be a down arrow
    button $w.b -image comboboxbuttonimage \
	    -command [list combobox_popup $w]
    pack $w.b -side right -expand 0 -fill both 
    pack $w.e -side left -fill both -expand 1

    ## Removable List Box
    toplevel $w.btm -cursor arrow
    wm withdraw $w.btm
    wm overrideredirect $w.btm 1
    wm transient $w.btm [winfo toplevel $w]
  
    listbox $w.btm.lbox -bd 2 -relief sunken -width 5 -height 5 \
	    -yscrollcommand [list $w.btm.vs set] -selectmode browse 
    scrollbar $w.btm.vs -width 10 -bd 1 -orient vertical \
		-command [list $w.btm.lbox yview]
    pack $w.btm.lbox -side left -fill both -expand 1

    bindtags $w.e [linsert [bindtags $w.e] 1 ComboEntry]
    bindtags $w.btm.lbox [linsert [bindtags $w.btm.lbox] 1 ComboList]

    return $w
}

proc combobox_setlist {w args} {
    if {[winfo class $w] != "ComboBox"} {
	return -code error "$w not a ComboBox"
    }
    if [winfo ismapped $w.btm] {
	wm withdraw $w.btm
	$w.b configure -relief raised
    }
    $w.btm.lbox delete 0 end
    eval $w.btm.lbox insert end $args
}

proc combobox_getlist w {
    if {[winfo class $w] != "ComboBox"} {
	return -code error "$w not a ComboBox"
    }
    return [$w.btm.lbox get 0 end]
}    

proc combobox_popup w {
    if {[winfo class $w] != "ComboBox"} {
	return -code error "$w not a ComboBox"
    }
    if [winfo ismapped $w.btm] {
	wm withdraw $w.btm
	$w.b configure -relief raised
    } else {
	wm geometry $w.btm [winfo width $w.e]x[winfo reqheight $w.btm]+[winfo rootx $w]+[expr [winfo rooty $w]+[winfo reqheight $w]]
	if {[$w.btm.lbox size] > 5} {
	    pack $w.btm.vs -side right -fill y
	} elseif {[winfo manager $w.btm.vs] == "pack"} {
	    pack forget $w.btm.vs
	}
	wm deiconify $w.btm
	raise $w.btm
	$w.b configure -relief sunken
	focus $w.btm.lbox
    }
}

proc combobox_find W {
    set found 0
    set tmp [$W get]
    set lbox [winfo parent $W].btm.lbox
    foreach item [$lbox get 0 end] {
	if ![string first $tmp $item] {
	    incr found
	    set match $item
	}
    }
    switch $found {
	0 {}
	1 {
	    $W delete 0 end
	    $W insert end $match
	}
	default {
	    bell
	}
    }
}
proc combobox_add W {
    set i 1
    set tmp [$W get]
    foreach item [[winfo parent $W].btm.lbox get 0 end] {
	if ![string compare $item $tmp] {
	    set i 0
	    break
	}
    }
    if $i {
	[winfo parent $W].btm.lbox insert end $tmp
    }
    wm withdraw [winfo parent $W].btm
}

proc combobox_get {W {y ""}} {
    set e [winfo parent [winfo parent $W]].e
    if [$W size] {
	$e delete 0 end	
		if {$y != ""} {
			$e insert end [$W get [$W nearest $y]]
    	} else {
			if {[$W curselection] != ""} {
				$e insert end [$W get [$W curselection]]
			}
		}
	}
    wm withdraw [winfo parent $W]
    [winfo parent $e].b configure -relief raised
	focus $e
}

#### END COMBOBOX CODE ------8<------------8<---------8<-------------


